@@ -1,3 +1,4 @@
+# $Id: Rule.pm,v 1.2 2003/06/19 02:57:47 grholt Exp $
 use strict qw(vars subs);
 
 package Rule;
@@ -144,9 +145,25 @@ sub find_all_targets_dependencies {
 # Now expand the command string.  This must be done last because it
 # can depend on both the explicit targets and the explicit dependencies.
 #
-  my $command_string = $makefile->expand_text($self->{COMMAND_STRING},
-					      $self->{RULE_SOURCE});
+  my $command_string;
+  if (@explicit_targets && $explicit_targets[0]->{TARGET_SPECIFIC_VARS})
+  {
+    local $Makefile::target_specific = $explicit_targets[0]->{TARGET_SPECIFIC_VARS};
+                                # Temporarily set up target-specific variables,
+                                # if there actually are any.
+    local $Makefile::target_specific_reexpand = $explicit_targets[0]->{TARGET_SPECIFIC_REEXPAND};
+                                # Remember the difference between = and :=
+                                # variables for target-specific vars too.
+    $command_string = $makefile->expand_text($self->{COMMAND_STRING},
+                                             $self->{RULE_SOURCE});
 				# Get the text of the command.
+  }                             # Turn off the target-specific variables.
+  else {
+    $command_string = $makefile->expand_text($self->{COMMAND_STRING},
+                                             $self->{RULE_SOURCE});
+				# Get the text of the command.
+  }
+
   $command_string =~ s/^\s+//;  # Strip out leading and trailing whitespace
   $command_string =~ s/\s+$//;  # so we don't trigger unnecessary rebuilds
                                 # quite as often.
@@ -542,7 +559,7 @@ or
 	    unless $scanner_warnings{$self->source}++;
 	}
       }
-    }	
+    }
   }
 }
 
