@@ -1,9 +1,11 @@
+# $Id: TextSubs.pm,v 1.2 2003/07/18 21:11:08 grholt Exp $
 package TextSubs;
 require Exporter;
 @ISA = qw(Exporter);
 
 @EXPORT = qw(index_ignoring_quotes split_on_whitespace split_on_colon unquote
-	     requote format_exec_args whitespace_len hash_neq);
+	     requote format_exec_args whitespace_len hash_neq
+             is_cpp_source_name is_object_or_library_name);
 
 #
 # This module contains a few subroutines for manipulating text, mostly for
@@ -586,4 +588,30 @@ sub hash_neq {
   return '';			# No difference.
 }
 
+=head2 is_cpp_source_name
+
+  if (is_cpp_source_name($filename))  { ... }
+
+Returns true if the given filename has the appropriate extension to be
+a C or C++ source or include file.
+
+=cut
+
+sub is_cpp_source_name {
+  return $_[0] =~ /\.(?:[ch]|cc|hh|[ch]xx|[ch]pp|[ch]\+\+|moc|x[bp]m)$/i;
+}
+
+=head2 is_object_name
+
+  if (is_object_or_library_name($filename)) { ... }
+
+Returns true if the given filename has the appropriate extension to be some
+sort of object or library file.
+
+=cut
+
+sub is_object_or_library_name {
+  return $_[0] =~ /\.(?:l[ao]|[ao]|s[aol](?:\.[\d\.]+))$/;
+}
+
 1;
